import importlib
import pkgutil
import numpy as np
import warnings
from scipy import sparse
from scipy.special import gamma


from .lmcpcpLink import libmonteCarloPCP
from . import enums
from . import packingMixins

import os
import time


# Constructs a dictionary that dynamically loads all mixin objects in
# packingMixins
mixins = {}
for loader, moduleName, isPackage in pkgutil.walk_packages(
		packingMixins.__path__
):
	# Imports the module containing the Mixin object
	module = importlib.import_module(
		__name__.split(".")[0] + ".packingMixins." + moduleName,
	)
	# Check if the mixin module has the ENABLED flag, and if it does, appends
	# appends its Mixin to the dictionary
	mixins[moduleName] = getattr(module, "Mixin")

class MCPacking(libmonteCarloPCP.MCPacking, *mixins.values()):
	def __init__(
		self,
		nDim=None,
		numParticles=None,
		phi=None
		):
		libmonteCarloPCP.MCPacking.__init__(self,nDim,numParticles,phi)

	# Getters -------------------------------------------------------------
	def getBoxSize(self):
		return np.array(libmonteCarloPCP.MCPacking.getBoxSize(self))

	def getD9Shift(self):
		return np.array(libmonteCarloPCP.MCPacking.getD9Shift(self))		

	def getPositions(self):
		posData = np.array(libmonteCarloPCP.MCPacking.getPositions(self))
		# Reshape positions data into an appropriately shaped array
		return posData.reshape(
			(self.getNumParticles(), self.getNDim())
			)

	def _getLastPositions(self):
		posData = np.array(libmonteCarloPCP.MCPacking._getLastPositions(self))
		# Reshape positions data into an appropriately shaped array
		return posData.reshape(
			(self.getNumParticles(), self.getNDim())
			)

	def getConstraint(self):
		conData = np.array(libmonteCarloPCP.MCPacking.getConstraint(self))
		return conData.reshape(
			(self.getNumParticles(), self.getNDim())
			)
	def getExternalForce(self):
		externalForceData = np.array(libmonteCarloPCP.MCPacking.getExternalForce(self))
		return externalForceData.reshape(
			(self.getNumParticles(), self.getNDim())
		)

	def getMKLambdas(self):
		lambdaData = np.array(libmonteCarloPCP.MCPacking.getMKLambdas(self))
		return lambdaData.reshape(
			(self.getNumParticles(), self.getNumParticles(), self.getNDim())
			)

	def pbcDistance(self,pos1,pos2):
		return libmonteCarloPCP.MCPacking.pbcDistance(self,pos1,pos2)

	def pbcDisplacement(self,pos1,pos2):
		return np.array(libmonteCarloPCP.MCPacking.pbcDisplacement(self,pos1,pos2))

	def pbcParticleDistance(self,p1,p2):
		return libmonteCarloPCP.MCPacking.pbcParticleDistance(self,p1,p2)

	def pbcParticleDisplacement(self,p1,p2):
		return np.array(libmonteCarloPCP.MCPacking.pbcParticleDisplacement(self,p1,p2))

	def getNetForces(self):
		forceData = np.array(libmonteCarloPCP.MCPacking.getNetForces(self))
		# Reshape positions data into an appropriately shaped array
		return forceData.reshape(
			(self.getNumParticles(), self.getNDim())
			)

	def getTotalDisplacement(self):
		dispData = np.array(libmonteCarloPCP.MCPacking.getTotalDisplacement(self))
		return dispData.reshape(
			(self.getNumParticles(), self.getNDim())
			)

	def getUnwrappedPositions(self):
		posData = np.array(libmonteCarloPCP.MCPacking.getUnwrappedPositions(self))
		# Reshape positions data into an appropriately shaped array
		return posData.reshape(
			(self.getNumParticles(), self.getNDim())
			)

	def getCrystalPositions(self):
		posData = np.array(libmonteCarloPCP.MCPacking.getCrystalPositions(self))
		# Reshape positions data into an appropriately shaped array
		return posData.reshape(
			(self.getNumParticles(), self.getNDim())
			)

	def getCenterOfMassShift(self):
		return np.array(libmonteCarloPCP.MCPacking.getCenterOfMassShift(self))

	def getRadii(self):
		return np.array(libmonteCarloPCP.MCPacking.getRadii(self))

	def getMCMCBlockedDirection(self):
		return np.array(libmonteCarloPCP.MCPacking.getMCMCBlockedDirection(self))

	def uniqueParticles(self,cutoff):
		return np.array(libmonteCarloPCP.MCPacking.uniqueParticles(self,cutoff))

	# Setters -------------------------------------------------------------
	def setPositions(self,posArray):
		libmonteCarloPCP.MCPacking.setPositions(self, posArray.flatten())

	def _setLastPositions(self,posArray):
		libmonteCarloPCP.MCPacking._setLastPositions(self, posArray.flatten())

	def setConstraint(self,conArray):
		libmonteCarloPCP.MCPacking.setConstraint(self, conArray.flatten())

	def setExternalForce(self,externalForceArray):
		libmonteCarloPCP.MCPacking.setExternalForce(self, externalForceArray.flatten())

	def translate(self,stepSize,translationVector):
		libmonteCarloPCP.MCPacking.translate(self, stepSize, translationVector.flatten())

	def setMKLambdas(self,lambdaArray):
		libmonteCarloPCP.MCPacking.setMKLambdas(self, lambdaArray.flatten())

	def setMKStartPoint(self,startPos=None,startRadii=None,randomString=None):
		if (startPos is None):
			startPos = self.getUnwrappedPositions()
		if (startRadii is None):
			startRadii = self.getRadii()
		if (randomString is None):
			randomString = self.getMTState()
		libmonteCarloPCP.MCPacking.setMKStartPoint(self, startPos.flatten(), startRadii.flatten(), randomString)

	def getMKStartPoint(self):
		startPos, startRadii, randomPoint = libmonteCarloPCP.MCPacking.getMKStartPoint(self)
		return np.array(startPos).reshape((self.getNumParticles(), self.getNDim())), np.array(startRadii), randomPoint

	def setCrystalPositions(self,posArray):
		libmonteCarloPCP.MCPacking.setCrystalPositions(self, posArray.flatten())		

	def setRadii(self,radArray):
		libmonteCarloPCP.MCPacking.setRadii(self, radArray.flatten())

	def setMCMCBlockedDirection(self,dirArray):
		libmonteCarloPCP.MCPacking.setMCMCBlockedDirection(self, dirArray.flatten())

	def createD9Plus(self,cryArray):
		cryPosData = np.array(libmonteCarloPCP.MCPacking.createD9Plus(self, cryArray.flatten()))
		return cryPosData.reshape(
			(self.getNumParticles(), self.getNDim())
			)

	def getRandomGaussianVector(self,cols,rows,mean=0,std=1):
		randomData = np.array(libmonteCarloPCP.MCPacking.getRandomGaussianVector(self, cols*rows, mean, std))
		return randomData.reshape(cols,rows)

	def getRandomVector(self,cols,rows,maxElement=1):
		randomData = np.array(libmonteCarloPCP.MCPacking.getRandomVector(self, cols*rows, maxElement))
		return randomData.reshape(cols,rows)

	def setBiRadii(self, sizeRatio = 1.4, numberRatio = 0.5):
		phi = self.getPhi()
		nParticles = self.getNumParticles()
		radii = np.ones(nParticles)
		radii[int(nParticles*numberRatio):] = sizeRatio
		self.setRadii(radii)
		self.setPhi(phi)

	def setMonoRadii(self):
		phi = self.getPhi()
		self.setRadii(np.ones(self.getNumParticles()))
		self.setPhi(phi)

	def calcMCOverlap(self,initialPosArray):
		return libmonteCarloPCP.MCPacking.calcMCOverlap(self, initialPosArray.flatten())

	def getMinRadius(self):
		return np.min(self.getRadii())

	def getMaxRadius(self):
		return np.max(self.getRadii())

	def getNeighbors(self):
		"""Return the neighbor lists as a sparse boolean array."""
		indices, indptr = libmonteCarloPCP.MCPacking.getNeighbors(self)
		return sparse.csr_matrix(
			(np.ones(len(indices), dtype=np.bool), indices, indptr),
			shape=(self.getNumParticles(),) * 2,
			dtype=np.bool
			)

	def getContacts(self,gap=0.0):
		"""Return the neighbor lists as a sparse boolean array."""
		indices, indptr = libmonteCarloPCP.MCPacking.getContacts(self,gap)
		return sparse.csr_matrix(
			(np.ones(len(indices), dtype=np.bool), indices, indptr),
			shape=(self.getNumParticles(),) * 2,
			dtype=np.bool
			)

	def getPerParticleStressTensor(self):
		"""Return the stress tensor (d by d matrix) for each particle as a
		numpy array of numParticle elements"""
		stData = np.array(libmonteCarloPCP.MCPacking.getPerParticleStressTensor(self))
		return stData.reshape(
			(self.getNumParticles(), self.getNDim(), self.getNDim())
			)

	def getStressTensor(self):
		"""
		Return the stress tensor for the entire system as a single
		d by d matrix.
		The stress tensor is defined as:
		sigma_{alpha beta} ~ -phi / N sum_k f_k^alpha r_k^beta
		Where alpha and beta are indicies over dimension and k is an index
		over contacts, f_k is the force between a pair of contacting particles
		and r_k is the distance between said particles.
		"""
		perParticleStress = self.getPerParticleStressTensor()
		return np.mean(perParticleStress, axis=0)

	def getPressure(self):
		stress = self.getStressTensor()
		# p = -Tr(S)/d
		return -np.trace(stress)/self.getNDim()

	def getPerParticlePressure(self):
		perParticleStress = self.getPerParticleStressTensor()
		return -np.trace(perParticleStress, axis1=1, axis2=2)/self.getNDim()
	
	def getMinGap(self,maxSearchGap=1.0):
		return libmonteCarloPCP.MCPacking.getMinGap(self,maxSearchGap)

	def getContactVectors(self, particleMask = None, normalized = False, stable = False, gap = 0.0):
		"""
		generates the vectors between contacting particles

		args:
			particleMask: The list of particles under consideration (int)
			normalized: Whether or not we are getting normalized vectors (bool)
			stable: Whether or not to only return stable particles (bool)
				Note, this only is used if particleMask is not provided
		returns: The matrix of contact vectors (coo np.quad)
		"""
		# Extract the contacts from libpyCudaPacking
		# Contact vectors is returned as a block coordinate matrix with block
		# size (1, nDim)
		block1, block2, cv = libmonteCarloPCP.MCPacking.getContactVectors(self,gap)
		numParticles = self.getNumParticles()
		nDim = self.getNDim()
		block1 = np.array(block1,dtype=np.int64)
		block2 = np.array(block2,dtype=np.int64)
		cv = np.array(cv,dtype=np.float64)


		# Create the list of particles of interest
		particleMask = self.makeParticleMask(particleMask, stable)
		particleIndices = particleMask[block1]*particleMask[block2]

		# the block indices need to be expanded into overall indices
		i1 = np.repeat(block1[particleIndices], nDim)
		i2 = np.repeat(block2[particleIndices], nDim) * nDim
		i2 += np.arange(i2.size, dtype=np.int64) % nDim
		contactMatrix = sparse.coo_matrix(
			(cv.view(np.float64)[np.repeat(particleIndices,nDim)], (i1, i2)),
			shape = (numParticles, numParticles * nDim),
			dtype = np.float64,
			)
		if normalized:
			data = contactMatrix.data.reshape(contactMatrix.nnz//nDim, nDim)
			norms = np.sqrt(np.sum(data**2,axis=1))
			contactMatrix.data /= np.repeat(norms, nDim)
		return contactMatrix

	def getSmallGaps(self, maxGap=0):
		indices1, indices2, gaps = libmonteCarloPCP.MCPacking.getSmallGaps(self, maxGap)
		return sparse.coo_matrix(
			(gaps, (indices1, indices2)),
			shape=(self.getNumParticles(),self.getNumParticles())
		)

	def getSmallOverlaps(self, maxOverlap=0):
		indices1, indices2, overlaps = libmonteCarloPCP.MCPacking.getSmallOverlaps(self, maxOverlap)
		return sparse.coo_matrix(
			(overlaps, (indices1, indices2)),
			shape=(self.getNumParticles(),self.getNumParticles())
		)

	def setInversePower(self, inversePower = 12, inversePowerCutoff = 1.25):
		# inversePowerC0 = np.quad('-1.924')
		# inversePowerC1 = np.quad('2.111')
		# inversePowerC2 = np.quad('-0.59109')
		inversePowerC0 = -(inversePower+4.0)*(inversePower+2.0)/(8.0*(inversePowerCutoff**inversePower))
		inversePowerC1 = inversePower*(inversePower+4.0)/(4.0*(inversePowerCutoff**(inversePower+2.0)))
		inversePowerC2 = -inversePower*(inversePower+2.0)/(8.0*(inversePowerCutoff**(inversePower+4.0)))
		libmonteCarloPCP.MCPacking.setInversePowerConstants(
			self, inversePowerCutoff, inversePowerC0, inversePowerC1, inversePowerC2
			)
		libmonteCarloPCP.MCPacking.setInversePower(self, inversePower)

	# Other functions ----------------------------------------------------
	# def save(self,basedir):
	# 	# Here, it is easier to create directories in python, so we create
	# 	# basedir, and then use the save function of mc.

	# 	# TODO: Just do this all in python
	# 	os.makedirs(basedir,exist_ok=True)
	# 	libmonteCarloPCP.MCPacking.save(self, basedir)

	def initFIREMinimizer(
			self,
			a_start=0.1, f_dec=0.5, f_inc=1.1, f_a=0.99, dt=0.001, dt_max=0.1, 
			fire_alpha=0.1, maxForceRescale=0.1, N_min=5, lastRestart=0
			):
		FIREParams = np.array([a_start,f_dec,f_inc,f_a,dt,dt_max,fire_alpha])
		libmonteCarloPCP.MCPacking.initFIREMinimizer(
			self,FIREParams,maxForceRescale,N_min,lastRestart
			)

	def initSDMinimizer(self,initialStepSize=1e-3, maxForceRescale=0.1, f_stepSizeDecrease=0.5, 
		checkGradient=True, gradientCheck=0.0, recheckRate=5, sdLastDecrease=0, sdLastIncrease=0, time=0.0):
		libmonteCarloPCP.MCPacking.initSDMinimizer(self,initialStepSize,maxForceRescale,f_stepSizeDecrease,
			checkGradient,gradientCheck,recheckRate,sdLastDecrease,sdLastIncrease,time)

	def removeParticles(self, toRemove = None, numParticles = 0):
		if numParticles > self.getNumParticles():
			print("Cannot remove more particles than exist")
		else:
			if (toRemove is None):
				toRemove = -1 * np.ones(numParticles,dtype=int)
				for i in range(numParticles):
					while(toRemove[i] == -1 or len(toRemove[:(i+1)]) != len(np.unique(toRemove[:(i+1)]))):
						toRemove[i] = int(np.floor(self.getRandomVector(1,1,self.getNumParticles())[0]))
			# print(toRemove)
			toRemove[::-1].sort()
			# print(toRemove)
			if ((toRemove>self.getNumParticles()).any() or (toRemove<0).any()):
				print("Cannot delete particles which do not exist")
			elif not np.issubdtype(toRemove.dtype,np.integer):
				print("Particle numbers must be integers")
			else:
				libmonteCarloPCP.MCPacking.removeParticles(self,toRemove)		

	def makeParticleMask(self, particleMask = None, stable = False):
		"""
		Create a particle mask for use in identifying particles of interest.
		The default behavior is to return a mask numParticle True values.

		args:
			particleMask (bool list): If set, return this mask
			stable (bool): Switch to select out only the stable particles, if
				particleMask is not set already
		"""
		if particleMask is None:
			if stable:
				particleMask = self.getStableListAndExcess()[0]
			else:
				particleMask = np.array([True]*self.getNumParticles())
		return particleMask

	def hyperBallCoefficient(self,dim):
		return (np.pi**(dim/2.0))/gamma(dim/2.0 + 1)

	def generateGCF(self,xi,fourierCutoff=20):
		# Must refactor so that it can be done in arbitrary dimensions
		pos = self.getPositions()
		c = 0*pos
		for d in range(self.getNDim()):
			A = self.getRandomGaussianVector(fourierCutoff,fourierCutoff,0,1/(2*np.pi));
			B = self.getRandomVector(fourierCutoff,fourierCutoff,2*np.pi);
			for pID in range(self.getNumParticles()):
				for n in range(fourierCutoff):
					for m in range(fourierCutoff):
						c[pID,d] = c[pID,d] + 2*np.pi*np.sqrt(2)*xi*A[n,m]*np.exp(-4*np.pi*np.pi*(n*n + m*m)*xi*xi/8)*np.cos(B[n,m] + 2*np.pi*n*pos[pID,0] - 2*np.pi*m*pos[pID,1])
		c -= np.mean(c,0)
		c /= np.sqrt(np.sum(c**2))
		return c

	# def gofr(self,pos,rad,numSteps,maxRad = None,minRad = None,neighbors = None, numVolumePoints = 1000000):
	# 	nParticles = pos.shape[0]
	# 	dim = pos.shape[1]
	# 	if (maxRad is None or maxRad > np.sqrt(dim)/2):
	# 		maxRad = np.sqrt(dim)/2
	# 	if (minRad is None or minRad < 2.0 * rad):
	# 		minRad = 2.0 * rad
	# 	stepSize = (maxRad - minRad)/numSteps
	# 	r = 2.0*rad + (np.arange(numSteps) + 0.5)*stepSize
	# 	outer = r + stepSize/2
	# 	inner = r - stepSize/2
	# 	g = np.zeros(numSteps)
	# 	volumeSlice = np.zeros(numSteps)
	# 	points = np.random.rand(numVolumePoints,dim) * self.getBoxSize()
	# 	pointDistances = np.sum((points - self.getBoxSize()/2.0)**2,1)
	# 	for i in range(numSteps):
	# 		# r[i] = 2.0*rad + (i + 0.5)*stepSize
	# 		# outer[i] = r[i] + stepSize/2
	# 		# inner = r[i] - stepSize/2
	# 		if outer[i] < np.min(self.getBoxSize())/2.0:
	# 			volumeSlice[i] = (outer[i]**dim - inner[i]**dim) * self.hyperBallCoefficient(dim)
	# 		else:
	# 			volumeSlice[i] = np.sum(np.logical_and(pointDistances < outer[i]**2,pointDistances > inner[i]**2))/numVolumePoints
	# 	for i in range(nParticles):
	# 		if neighbors is None:
	# 			if (i == nParticles - 1):
	# 				break
	# 			jRange = range(i+1,nParticles)
	# 		else:
	# 			jRange = neighbors.indices[neighbors.indptr[i]:neighbors.indptr[i+1]]
	# 		for j in jRange:
	# 			distance = self.pbcDistance(pos[i],pos[j])
	# 			if distance < maxRad:
	# 				g[int(np.floor((distance - 2.0*rad)/stepSize))] += 1 
	# 	for i in range(numSteps):
	# 		if neighbors is None:
	# 			g[i] /= (nParticles * nParticles * volumeSlice[i] * 0.5)
	# 		else:
	# 			g[i] /= (nParticles * nParticles * volumeSlice[i])
	# 	return g, r

	def gofrCummulative(self,pos,rad,numSteps,maxRad = None,minRad = None, neighbors = None, logBin = False,binRatio=100):
		nParticles = pos.shape[0]
		dim = pos.shape[1]
		if (maxRad is None or maxRad > np.sqrt(dim)/2):
			maxRad = np.sqrt(dim)/2
		if (minRad is None or minRad < 2.0 * rad):
			minRad = 2.0 * rad
		if (logBin == True):
			bins = np.append(np.array([1.0]),np.logspace(np.log10(maxRad/rad - 1)-4,np.log10(maxRad/rad - 1),num = numSteps)+1.0)*2.0*rad
			r = np.sqrt(bins[0:-1]*bins[1:])
		else:
			bins = np.linspace(minRad,maxRad,num = numSteps + 1)
			r = (bins[1:] + bins[0:-1])/2
		g = np.zeros(numSteps)
		dist = np.zeros(int(nParticles*(nParticles-1)/2))
		ii = 0
		for i in range(nParticles-1):
			for j in range(i+1,nParticles):
				dist[ii] = self.pbcDistance(pos[i],pos[j])
				ii+=1
		dist.sort()
		for i in range(numSteps):
			g[i] = 2*np.searchsorted(dist, r[i], side="left")
		return g, r

	def gofr(self,numSteps,maxRad = None,minRad = None, numVolumePoints = 1000000,logBin = False,logBinMin = -3, logBinMax = 0):
		rad = np.min(self.getRadii())
		dim = self.getNDim()
		if (maxRad is None or maxRad > np.sqrt(dim)/2):
			if self.getGeometryType() == enums.geometryEnum.E8:
				maxRad = 1.0
			else:
				maxRad = np.sqrt(dim)/2
		if (minRad is None):
			minRad = 2.0 * rad
		if (logBin == True):
			bins = np.append(np.array([1.0]),np.logspace(np.log10(maxRad/(2*rad) - 1)+logBinMin,np.log10(maxRad/(2*rad) - 1)+logBinMax,num = numSteps)+1.0)*2.0*rad
			r = np.sqrt(bins[0:-1]*bins[1:])
		else:
			bins = np.linspace(minRad,maxRad,num = numSteps + 1)
			r = (bins[1:] + bins[0:-1])/2
		g = self.calcGofR(bins,numVolumePoints)
		return np.array(g),np.array(r)



	# def gofr(self,pos,rad,numSteps,maxRad = None,minRad = None, neighbors = None, numVolumePoints = 1000000,logBin = False,binRatio=100):
		# nParticles = pos.shape[0]
		# dim = pos.shape[1]
		# if (maxRad is None or maxRad > np.sqrt(dim)/2):
		# 	maxRad = np.sqrt(dim)/2
		# if (minRad is None or minRad < 2.0 * rad):
		# 	minRad = 2.0 * rad
		# if (logBin == True):
		# 	bins = np.append(np.array([1.0]),np.logspace(np.log10(maxRad/rad - 1)-4,np.log10(maxRad/rad - 1),num = numSteps)+1.0)*2.0*rad
		# 	r = np.sqrt(bins[0:-1]*bins[1:])
		# else:
		# 	bins = np.linspace(minRad,maxRad,num = numSteps + 1)
		# 	r = (bins[1:] + bins[0:-1])/2
		# g = np.zeros(numSteps)
		# volumeSlice = (bins[1:]**dim - bins[0:-1]**dim) * self.hyperBallCoefficient(dim)
		# if bins[-1] > np.min(self.getBoxSize())/2.0:
		# 	approxVolumeSlice = np.zeros(numSteps)
		# 	maxSetSize = 1000000
		# 	for ii in range(int(np.ceil(numVolumePoints/maxSetSize))):
		# 		points = np.random.rand(maxSetSize,dim)*self.getBoxSize()
		# 		pointDistance = np.sum((points - self.getBoxSize()/2.0)**2,1)
		# 		place = np.digitize(pointDistance,bins**2)
		# 		for i in range(maxSetSize):
		# 			if (place[i] <= numSteps and place[i] > 0):
		# 				approxVolumeSlice[place[i]-1] += 1
		# 	approxVolumeSlice /= numVolumePoints
		# 	volumeSlice[bins[1:] > np.min(self.getBoxSize())/2.0] = approxVolumeSlice[bins[1:] > np.min(self.getBoxSize())/2.0]
		# for i in range(nParticles):
		# 	if neighbors is None:
		# 		if (i == nParticles - 1):
		# 			break
		# 		jRange = range(i+1,nParticles)
		# 	else:
		# 		jRange = neighbors.indices[neighbors.indptr[i]:neighbors.indptr[i+1]]
		# 	for j in jRange:
		# 		distance = self.pbcDistance(pos[i],pos[j])
		# 		place = np.digitize(np.array([distance]),bins)[0]
		# 		if (place < numSteps and place > 0):
		# 			g[place] += 1 
		# for i in range(numSteps):
		# 	if neighbors is None:
		# 		g[i] /= (nParticles * nParticles * volumeSlice[i] * 0.5)
		# 	else:
		# 		g[i] /= (nParticles * nParticles * volumeSlice[i])
		# return g, r

