import os
import numpy as np
from .. import enums
# from .. import fileIO
# import time
import warnings
import shutil
import time

class Mixin():
	def neighborRecalc(
		self,
		*args,
		cutDistance=1.0,
		cutRescale=0.5,
		minCut=0.1,
		iteration=0,
		lastRecalc=0,
		scaleFreq=16,
		**kwargs
		):
		minRadius = np.min(self.getRadii())
		minCut = minCut
		if self.getPotentialType() == enums.potentialEnum.log:
			if (iteration == 0 or self.getSumOfMaxDisplacements() >= 2.0 * cutDistance * minRadius):
				self.calcNeighborsCut(cutDistance)
				lastRecalc = iteration
				self.resetDisplacement()
		else:
			if (iteration == 0 or self.getSumOfMaxDisplacements() >= 2.0 * cutDistance * minRadius):
			# if (iteration == 0 or 3.0 * self.getMaxDisplacement() > cutDistance * minRadius):
				self.calcNeighborsCut(cutDistance)
				# print('recalculating ' + str(iteration))
				# print(cutDistance)
				# print(scaleFreq)
				lastRecalc = iteration
				self.resetDisplacement()
			elif (iteration - lastRecalc == scaleFreq and cutDistance > minCut):
				cutDistance *= cutRescale
				if cutDistance < minCut:
					cutDistance = minCut
				self.calcNeighborsCut(cutDistance)
				lastRecalc = iteration
				self.resetDisplacement()
		return lastRecalc, cutDistance

	def minimizeFIRE(self, *args, **kwargs):
		"""
		Runs a single FIRE minimization. See FIREMinimizer for further
		documentation
		"""
		iteration = 0
		for iteration in self.FIREMinimizer(*args, **kwargs):
			pass
		return iteration

	def FIREMinimizer(self, *args, checkpointDir = None, saveFreq = 3600, removeCheckpointDir = True, **kwargs):
		"""
		Creates a generator that runs a single FIRE minimization

		args:
			checkpointDir: The directory where all of the FIRE parameters
				and packing parameters are saved
			saveFreq: The time between saving (not a frequency)
		returns:
			iteration (int) number of FIRE steps
		"""
		params = self.initFIRE(*args, checkpointDir = checkpointDir, **kwargs)
		criticalForce = params["criticalForce"]
		maxIterations = params["maxIterations"]
		recalcNeighbors = params["recalcNeighbors"]
		dt_max = params["dt_max"]
		cutDistance = params["cutDistance"]
		cutRescale = params["cutRescale"]
		minCut = params["minCut"]
		iteration = params["iteration"]
		lastRecalc = params["lastRecalc"]
		scaleFreq = params["scaleFreq"]
		dt = params["dt"]
		fire_alpha = params["fire_alpha"]
		FIREIteration = params["FIREIteration"]
		prevTime = time.time() - saveFreq
		# continueFlag = True
		# lastPos = self.getPositions()
		# if self.getPotentialType() == enums.potentialEnum.log:
		# 	continueFlag = len(self.getSmallOverlaps().data) == 0
		while (
				self.getMaxUnbalancedForce() >= criticalForce and
				iteration != maxIterations #and continueFlag
				):
			if recalcNeighbors:
				# TODO: This cutDistance is only good for the contact potential
				lastRecalc, cutDistance = self.neighborRecalc(
					cutDistance=cutDistance,
					cutRescale=cutRescale,
					minCut=minCut,
					iteration=iteration,
					lastRecalc=lastRecalc,
					scaleFreq=scaleFreq)
			# lastPos = self.getPositions()
			self.minimizerLoopFIRE()
			# TODO: Decide whether the first iteration should return a 0 or a 1
			iteration += 1
			# if self.getPotentialType() == enums.potentialEnum.log:
			# 	continueFlag = len(self.getSmallOverlaps().data) == 0
			if checkpointDir is not None and time.time() - prevTime > saveFreq:
				prevTime = time.time()
				# TODO: This should also go into a function
				# update the dictionary
				params = self._updateFIRERestoreData(params, cutDistance,
					iteration, lastRecalc)
				self._updateFIREState(params,checkpointDir = checkpointDir)
			yield iteration
		# if not continueFlag:
		# 	self.setPositions(lastPos)
		if removeCheckpointDir and checkpointDir is not None:
			try:
				shutil.rmtree(checkpointDir)
			except FileNotFoundError:
				pass

	def _updateFIRERestoreData(self, params, cutDistance, iteration, lastRecalc):
		params["cutDistance"] = cutDistance
		params["iteration"] = iteration
		params["lastRecalc"] = lastRecalc
		params["dt"] = self.getFIRETimeStep()
		params["fire_alpha"] = self.getFIREDamping()
		params["FIREIteration"] = self.getFIREIteration()
		return params

	def initFIRE(
		self,
		criticalForce=-1,
		maxIterations=-1,
		recalcNeighbors=True,
		initialCut=1.0,
		minCut=0.1,
		cutRescale=0.5,
		cutDistance = 1,
		scaleFreq=16,
		maxForceRescale=0.1,
		dt_max=0.1,
		dt=0.001,
		fire_alpha=0.1,
		checkpointDir = None,
		*args,**kwargs
	   ):
		"""
		Initializes the FIRE minimization

		args:
			criticalExcessForce(str castable numerical type): maximum net force
				magnitude at which minimization is stopped, scaled to particle
				size
			maxIterations (int): maximum number of iterations, default -1,
				which will never trigger
			recalcNeighbors (bool): when set to True, neighbors are
				periodically recalculated useing a distance cutoff
			initialCut (np.quad): initial value for the distance cutoff, with
				units of scaled overlap
			minCut (np.quad): minimum value for the distance cutoff, with
				units of scaled overlap
			cutRescale (np.quad): factor by which the distance cutoff is
				rescaled when above minCut.
			scaleFreq (int): number of iterations without recalculating
				neighbors after which the distance cut is rescaled
			checkpointDir (str): where the packing will be loaded (full path)

		returns:
			the parameters for minimization in a dictionary
		"""
		try:
			params = self._loadFIRERestoreData(checkpointDir)
			dt_max = params["dt_max"]
			dt = params["dt"]
			fire_alpha = params["fire_alpha"]
			FIREIteration = params["FIREIteration"]
			self.initFIREMinimizer(dt_max=dt_max,dt=dt,fire_alpha=fire_alpha,
				lastRestart=FIREIteration,maxForceRescale = maxForceRescale)
			self.load(checkpointDir,["forces","velocities","lastPositions",
				"neighbors"])
			return params
		except (TypeError,FileNotFoundError):
			self.initFIREMinimizer(dt_max=dt_max,maxForceRescale=maxForceRescale,dt=dt,**kwargs)
			criticalForce = criticalForce / self.getMeanRadius()
			paramsNames = ["criticalForce","maxIterations","recalcNeighbors",
				"dt_max","cutDistance","cutRescale","minCut","iteration",
				"lastRecalc","scaleFreq","dt","fire_alpha",
				"FIREIteration"]
			paramsValues = [criticalForce,maxIterations,recalcNeighbors,dt_max,
				initialCut,cutRescale,minCut,0,0,scaleFreq,
				dt,fire_alpha,0]
			return dict(zip(paramsNames,paramsValues))

	def _loadFIRERestoreData(self, checkpointDir):
		params = {}
		with open(checkpointDir + "/FIRERestoreData.dat") as file:
			for line in file:
				name, scalarString = line.strip().split("\t")
				params[name] = scalarString
		intList = ["maxIterations","iteration","lastRecalc","scaleFreq",
			"FIREIteration"]
		boolList = ["recalcNeighbors"]
		for key in params:
			if key in intList:
				params[key] = int(np.float128(params[key]))
			elif key in boolList:
				params[key] = bool(params[key])
			else:
				params[key] = params[key]
		return params

	def _updateFIREState(self,params,checkpointDir = None):
		if not os.path.isdir(checkpointDir):
			self._saveFIREState(params,checkpointDir)
		try:
			os.rename(checkpointDir,checkpointDir + ".old")
		except OSError:
			warnings.warn("The checkpoint directory was not properly"
				" saved. You can try renaming the .old file that is here to"
				" the name of your checkpoint directory.")
			exit()
		self._saveFIREState(params,checkpointDir + ".new")
		os.rename(checkpointDir + ".new",checkpointDir)
		shutil.rmtree(checkpointDir + ".old")

	def _saveFIREState(self,params,directory = None):
		self.save(directory,["forces","velocities",
			"lastPositions","neighbors"],overwrite=True)
		with open(directory + "/FIRERestoreData.dat","w") as cpd:
			for name,value in params.items():
				cpd.write(name + "\t" + str(value) + "\n")

	def SDMinimizer(self, *args, checkpointDir = None, 
		saveFreq = 3600, removeCheckpointDir = True, **kwargs):
		"""
		Creates a generator that runs a single Steepest Descent minimization
		args:
			criticalExcessForce(str castable numerical type): maximum net force
				magnitude at which minimization is stopped, scaled to particle
				size
			maxIterations (int): maximum number of iterations, default -1,
				which will never trigger
			recalcNeighbors (bool): when set to True, neighbors are
				periodically recalculated useing a distance cutoff
			initialCut (np.quad): initial value for the distance cutoff, with
				units of scaled overlap
			minCut (np.quad): minimum value for the distance cutoff, with
				units of scaled overlap
			cutRescale (np.quad): factor by which the distance cutoff is
				rescaled when above minCut.
			scaleFreq (int): number of iterations without recalculating
				neighbors after which the distance cut is rescaled
			f_stepSize (np.quad): factor by which the steepest descent step
				size will decrease if a step is not allowed
			checkGradient (bool): check if the graident condition is satisfied
				and lower the step size if it isn't
			gradientCheck (np.quad): Minimum change in curvature allowed for a
				step. Default value is zero, meaning that the minimization
				wont allow steps past the minimum
		returns:
			iteration (int) number of Steepest Descent steps
		"""
		params = self.initSD(*args, checkpointDir = checkpointDir, **kwargs)
		criticalForce = params["criticalForce"]
		maxIterations = params["maxIterations"]
		recalcNeighbors = params["recalcNeighbors"]
		cutDistance = params["cutDistance"]
		cutRescale = params["cutRescale"]
		minCut = params["minCut"]
		iteration = params["iteration"]
		lastRecalc = params["lastRecalc"]
		scaleFreq = params["scaleFreq"]
		initialStepSize = params["initialStepSize"]
		f_stepSizeDecrease = params["f_stepSizeDecrease"]
		checkGradient = params["checkGradient"]
		gradientCheck = params["gradientCheck"]
		recheckRate = params["recheckRate"]
		sdLastIncrease = params["sdLastIncrease"]
		sdLastDecrease = params["sdLastDecrease"]
		sdTime = params["sdTime"]
		prevTime = time.time() - saveFreq
		self.initSDMinimizer(initialStepSize=initialStepSize,
			f_stepSizeDecrease=f_stepSizeDecrease, checkGradient=checkGradient,
			gradientCheck=gradientCheck, recheckRate=recheckRate,
			sdLastDecrease=sdLastDecrease, sdLastIncrease=sdLastIncrease, time=sdTime)
		# print(self.getMaxUnbalancedForce())
		# Scale the criticalForce to the meanRadius
		# criticalForce = criticalForce / self.getMeanRadius()
		# cutDistance = initialCut
		# iteration = 0
		# lastRecalc = 0

		if recalcNeighbors:
			lastRecalc, cutDistance = self.neighborRecalc(
					cutDistance=cutDistance,
					cutRescale=cutRescale,
					minCut=minCut,
					iteration=iteration,
					lastRecalc=lastRecalc,
					scaleFreq=scaleFreq)

		while (
				self.getMaxUnbalancedForce() >= criticalForce and
				iteration != maxIterations
				):
			# stepSize = initialStepSize
			# goodStep = False
			# lastPosition = self.getUnwrappedPositions()
			# lastForce = self.getNetForces()

			if recalcNeighbors:
				# TODO: This cutDistance is only good for the contact potential
				# if we're using the log we need something like
				# if 3 * self.getMaxDisplacement() > (cutDistance - cut0)
				lastRecalc, cutDistance = self.neighborRecalc(
					cutDistance=cutDistance,
					cutRescale=cutRescale,
					minCut=minCut,
					iteration=iteration,
					lastRecalc=lastRecalc,
					scaleFreq=scaleFreq)
						
			self.minimizerLoopSD()
			# print(self.getMaxUnbalancedForce())
			# print(self.getEnergy())
			iteration += 1
			if checkpointDir is not None and time.time() - prevTime > saveFreq:
				prevTime = time.time()
				# TODO: This should also go into a function
				# update the dictionary
				params = self._updateSDRestoreData(params, cutDistance,
					iteration, lastRecalc)
				self._updateSDState(params,checkpointDir = checkpointDir)
			yield iteration
		if removeCheckpointDir and checkpointDir is not None:
			try:
				shutil.rmtree(checkpointDir)
			except FileNotFoundError:
				pass

	def initSD(self,
		initialStepSize=1e-3,
		checkGradient = True,
		criticalForce=-1.0,
		maxIterations=-1.0,
		recalcNeighbors=True,
		initialCut=1.0,
		minCut=0.1,
		cutRescale=0.5,
		maxForceRescale=0.1,
		scaleFreq=16,
		f_stepSizeDecrease=0.5,
		gradientCheck = 0.0,
		recheckRate=5,
		checkpointDir = None,
		*args,**kwargs
	   ):
		"""
		Initializes the SD minimization

		args:
			criticalExcessForce(str castable numerical type): maximum net force
				magnitude at which minimization is stopped, scaled to particle
				size
			maxIterations (int): maximum number of iterations, default -1,
				which will never trigger
			recalcNeighbors (bool): when set to True, neighbors are
				periodically recalculated useing a distance cutoff
			initialCut (np.quad): initial value for the distance cutoff, with
				units of scaled overlap
			minCut (np.quad): minimum value for the distance cutoff, with
				units of scaled overlap
			cutRescale (np.quad): factor by which the distance cutoff is
				rescaled when above minCut.
			scaleFreq (int): number of iterations without recalculating
				neighbors after which the distance cut is rescaled
			checkpointDir (str): where the packing will be loaded (full path)

		returns:
			the parameters for minimization in a dictionary
		"""
		try:
			params = self._loadSDRestoreData(checkpointDir)
			initialStepSize = params["initialStepSize"]
			f_stepSizeDecrease = params["f_stepSizeDecrease"]
			checkGradient = params["checkGradient"]
			gradientCheck = params["gradientCheck"]
			recheckRate = params["recheckRate"]
			sdLastDecrease = params["sdLastDecrease"]
			sdLastIncrease = params["sdLastIncrease"]
			sdTime = params["sdTime"]
			self.initSDMinimizer(initialStepSize=initialStepSize,
				f_stepSizeDecrease=f_stepSizeDecrease, checkGradient=checkGradient,
				gradientCheck=gradientCheck, maxForceRescale=maxForceRescale, recheckRate=recheckRate,
				sdLastDecrease=sdLastDecrease,sdLastIncrease=sdLastIncrease,time=sdTime)
			self.load(checkpointDir,["lastPositions"])
			self.calcNeighborsCut(initialCut)
			self.calcForceEnergy()
			return params
		except (TypeError,FileNotFoundError):
			self.initSDMinimizer(initialStepSize=initialStepSize,
				f_stepSizeDecrease=f_stepSizeDecrease, checkGradient=checkGradient,
				gradientCheck=gradientCheck, maxForceRescale=maxForceRescale,
				recheckRate=recheckRate, **kwargs)
			criticalForce = criticalForce / self.getMeanRadius()
			paramsNames = ["initialStepSize","f_stepSizeDecrease","checkGradient","gradientCheck",
				"sdLastDecrease","sdLastIncrease",
				"recheckRate","criticalForce","maxIterations","recalcNeighbors",
				"cutDistance","cutRescale","minCut","iteration",
				"lastRecalc","scaleFreq","sdTime"]
			paramsValues = [initialStepSize,f_stepSizeDecrease,checkGradient,gradientCheck,
				0,0,
				recheckRate,criticalForce,maxIterations,recalcNeighbors,
				initialCut,cutRescale,minCut,0,
				0,scaleFreq,0.0]
			self.calcNeighborsCut(initialCut)
			self.calcForceEnergy()
			return dict(zip(paramsNames,paramsValues))

	def minimizeSD(self, *args, **kwargs):
		"""
		Runs a single Steepest Descent minimization. See SteepestDescentMinimizer for futrther
		documentation
		"""
		iteration = 0
		for iteration in self.SDMinimizer(*args, **kwargs):
			pass
		return iteration

	def _loadSDRestoreData(self, checkpointDir):
		params = {}
		with open(checkpointDir + "/SDRestoreData.dat") as file:
			for line in file:
				name, scalarString = line.strip().split("\t")
				params[name] = scalarString
		intList = ["maxIterations","iteration","lastRecalc","scaleFreq","recheckRate",
			"sdLastDecrease","sdLastIncrease"]
		boolList = ["recalcNeighbors","checkGradient"]
		for key in params:
			if key in intList:
				params[key] = int(np.float128(params[key]))
			elif key in boolList:
				params[key] = bool(params[key])
			else:
				params[key] = np.float128(params[key])
		return params

	def _updateSDState(self,params,checkpointDir = None):
		if not os.path.isdir(checkpointDir):
			self._saveSDState(params,checkpointDir)
		try:
			os.rename(checkpointDir,checkpointDir + ".old")
		except OSError:
			warnings.warn("The checkpoint directory was not properly"
				" saved. You can try renaming the .old file that is here to"
				" the name of your checkpoint directory.")
			exit()
		self._saveSDState(params,checkpointDir + ".new")
		os.rename(checkpointDir + ".new",checkpointDir)
		shutil.rmtree(checkpointDir + ".old")

	def _saveSDState(self,params,directory = None):
		self.save(directory,["lastPositions"],overwrite=True)
		with open(directory + "/SDRestoreData.dat","w") as cpd:
			for name,value in params.items():
				cpd.write(name + "\t" + str(value) + "\n")

	def _updateSDRestoreData(self, params, cutDistance, iteration, lastRecalc):
		params["cutDistance"] = cutDistance
		params["iteration"] = iteration
		params["lastRecalc"] = lastRecalc
		params["sdLastDecrease"] = self.getSDLastDecrease()
		params["sdLastIncrease"] = self.getSDLastIncrease()
		params["initialStepSize"] = self.getSDStepSize()
		params["sdTime"] = self.getSDTime()
		return params

	def isostaticitySearch(
			self,
			phiC,
			phiInitial,
			maxDeltaZ=0,
			maxRattlerFraction=0.5,
			nStepsPerDecadePhi=10,
			epsilon=None,
			criticalForce=1e-20,
			phiResolution=1e-10,
			minimizer="FIRE",
			checkpointDir=None,
			saveFreq=3600,
			removeCheckpointDir=True,
	):
		"""
		Finds a packing fraction from above such that the packing
		minimizes to a given number of excess cpntacts, using a power law
		ansatz that should be accurate for contact power potentials at
		low pressure.
		args:
			phiC (np.quad): Guess for the jamming fraction
			phiInitial (np.quad): Initial packing fraction. This
				should be high enough that the minimized deltaZ
				is higher than maxDeltaZ
			maxDeltaZ (int): maximum number of excess contacts to halt
				minimization
			maxRattlerFraction (np.quad): steps are automatically diregarded
				if the fraction of rattlers exceeds this number
			epsilon (np.quad): This parameter allows for a gradual approach to
				the target phi. When a new phi is calculated using the power
				law ansatz, the phi that is actually used is
				epsilon*phi_guess + (1-elsilon)*phi_current
			nStepsPerDecadePhi (np.quad): Alternative parameter to epsilon
			criticalForce (np.quad): Stop condition for minimization
				steps
			phiResolution (np.quad): min difference in phi at which iteration
				stops.
			minimizer (string): minimization protocol. Currently
				accepts "CG" and "FIRE", or a custom defined function accepting
				a Packing object as argument
		"""
		phiResolution = phiResolution
		if epsilon is not None:
			epsilon = epsilon
		else:
			epsilon = 1 - 10**(-1/nStepsPerDecadePhi)

		if minimizer == "FIRE":
			def runMinimizer(useCheckpoint=True):
				if useCheckpoint:
					self.minimizeFIRE(criticalForce, checkpointDir=checkpointDir, saveFreq=saveFreq, removeCheckpointDir=removeCheckpointDir)
				else:
					self.minimizeFIRE(criticalForce)
		# elif minimizer == "CG":
		#	 def runMinimizer():
		#		 self.minimizeCG(criticalForce)
		elif isinstance(minimizer, str):
			raise ValueError(
				"minimizer argument \"{}\" not understood. "
				"Values \"FIRE\", \"CG\" allowed"
				.format(minimizer)
			)
		else:
			runMinimizer = minimizer

		# Set up brackets
		potentialPower = self.getPotentialPower()
		phiC = phiC
		maxPhi = phiInitial
		currentPhi = maxPhi
		finished = False

		self.setPhi(currentPhi)
		runMinimizer(useCheckpoint=False)
		currentEnergy = self.getEnergy()

		while not finished:
			newPhi = epsilon * phiC + (1 - epsilon) * currentPhi
			self.setPhi(newPhi)
			runMinimizer()
			newEnergy = self.getEnergy()
			scaledEnergyRatio = (
				(newEnergy / currentEnergy) **
				(1 / potentialPower)
			)
			# Recalculate the estimated jamming fraction
			phiC = (
				(newPhi - scaledEnergyRatio * currentPhi) /
				(1 - scaledEnergyRatio)
			)
			stableList, excess = self.getStableListAndExcess()
			rattlerFraction = (
				1 - np.sum(stableList) /
				self.getNumParticles()
			)
			# Update the upper bound if this step undershoots
			if (
					rattlerFraction < maxRattlerFraction
					and excess > maxDeltaZ
					and currentPhi - newPhi > phiResolution
				):
				currentPhi, currentEnergy = newPhi, newEnergy
			elif (
					rattlerFraction < maxRattlerFraction
					and currentPhi - newPhi > phiResolution
				):
				finished = True
			else:
				print(
					"pyCudaPacking.Packing.isostaticitySearch Warning: "
					"DeltaZ has not converged. Returning current state."
				)
				finished = True
			yield stableList, excess, phiC

	def pressureSearch(
			self,
			targetPressure,
			phiC,
			phiInitial,
			criticalForce=1e-13,
			pressureResolution=1e-10,
			phiResolution=1e-10,
			minimizer="FIRE",
			ifBelow = None

	):
		"""
		Finds a packing fraction from above such that the packing
		minimizes to a given pressure, using a power law ansatz
		that should be accurate for contact power potentials at
		low pressure.
		args:
			targetPressure (np.quad): target pressure
			phiC (np.quad): Guess for the jamming fraction
			phiInitial (np.quad): Initial packing fraction. This
				should be high enough that the minimized pressure
				is higher than targetPressure
			criticalForce (np.quad): Stop condition for minimization
				steps
			pressureResolution (np.quad): max relative error in pressure
				to stop iteration
			phiResolution (np.quad): min difference in phi at which iteration
				stops
			minimizer (string): minimization protocol. Currently
				accepts "CG" and "FIRE", or a custom defined function accepting
				a Packing object as argument
			ifBelow (np.quad): if the initial given phi yields a pressure below
				targetPressure, we multiply phi by ifBelow until it yields a
				pressure above targetPressure
		yields:
			newPressure (np.quad)
			phiC (np.quad)
		"""

		# TODO: Refactor this into a function
		if minimizer == "FIRE":
			def runMinimizer():
				self.minimizeFIRE(criticalForce=criticalForce)
		elif minimizer == "CG":
			def runMinimizer():
				self.minimizeCG(criticalForce)
		elif isinstance(minimizer, str):
			raise ValueError(
				"minimizer argument \"{}\" not understood. "
				"Values \"FIRE\", \"CG\" allowed"
				.format(minimizer)
			)
		else:
			runMinimizer = minimizer


		# Simple type coercion
		potentialPower = self.getPotentialPower()
		pressureExponent = 1 / (potentialPower - 1)
		currentPhi = phiInitial
		

		# Jump to the specified starting density and minimize and save state
		self.setPhi(currentPhi)
		runMinimizer()
		currentPressure = self.getPressure()

		# Check to see that we start out above the target because we can only
		# search down
		if currentPressure < targetPressure:
			#the quad conversion will turn ifBelow into a nan
			if np.isnan(ifBelow) or ifBelow <= 1:
				raise RuntimeError("initial pressure below target - to avoid this error, consider including the optional argument ifBelow")

			print(
					"pyCudaPacking.Packing.pressureSearch Warning: "
					"Initial pressure is below target pressure. "
					"This may have undesirable results"
				)
			while currentPressure < targetPressure:
				#increase phi until we are above targetPressure
				currentPhi=currentPhi*ifBelow
				self.setPhi(currentPhi)
				runMinimizer()
				currentPressure=self.getPressure()

		currentPos = self.getPositions()



		# Pick a new density in the hope that it's correct
		# TODO: Use our knowledge of the scaling of pressure with deltaPhi
		# to make a better guess than just halving the distance
		testPhi = (currentPhi + phiC) / 2

		# Run an infinite loop with breaks to exit
		while True:
			# Proposing a new density and calculating the resulting pressure
			self.setPhi(testPhi)
			runMinimizer()
			testPressure = self.getPressure()

			# If our test system has a smaller pressure than our current then
			# we are in good shape, let's figure out a new phiC estimate
			if testPressure < currentPressure:
				scaledPressureRatio = (testPressure / currentPressure) ** pressureExponent
				testPhiC = (testPhi - scaledPressureRatio * currentPhi) / (1 - scaledPressureRatio)
				if testPhiC > 0 and testPhiC < currentPhi:
					phiC = testPhiC

			# Return the pressure and phiCEstimate
			yield testPressure, phiC

			# We are finished if the testPressure is close enough to target pressure
			pressureDifference = testPressure - targetPressure
			if (np.abs(pressureDifference / targetPressure) < pressureResolution):
				# The iterator is finished when we break out of the while loop
				break
			# Otherwise, if we are above the target pressure choose a new
			# testphi that is smaller and yield information
			elif testPressure > targetPressure:
				# Make the test system the current system
				currentPressure = testPressure
				currentPhi = testPhi
				currentPos = self.getPositions()
				# TODO: Print shouldn't be in here, just for debugging
				# Calculate the new test density as where we think the targetPressure
				# should be
				testPhi = phiC + (currentPhi - phiC) * (targetPressure / currentPressure) ** pressureExponent
			# Otherwise, the test system is below the target and we need to reject
			# the proposed system and pick a new test system that is at a higher
			# density
			else:
				# Reject the test and restore the currentPos
				self.setPositions(currentPos)
				self.setPhi(currentPhi)
				# TODO: Print shouldn't be in here, just for debugging
				# We know that currentPhi is above target, so just go halfway
				testPhi = (currentPhi + testPhi) / 2



			# Check if currentPhi and testPhi are too close.  If they are, then
			# we are finished
			if np.abs(currentPhi - testPhi) / np.mean((currentPhi,testPhi)) < phiResolution:
				print(
					"pyCudaPacking.Packing.pressureSearch Warning: "
					"Pressure is not within resolution of targetPressure,"
					"but the change in phi is below phiResolution."
					"The system is at the smallest available pressure above target"
					"This is probably wrong behavior."
				)
				break


