#ifndef PACKINGIMPLEMENTATION_H_
#define PACKINGIMPLEMENTATION_H_

#include "PIMPLClasses.h"
#include "MCPacking.h"
#include "enumTypes.h"
#include "defs.h"
#include <algorithm>
#include <numeric>
#include <vector>
#include <iterator>
#include <functional>
#include <iostream>
#include <cmath>

//Implementation

using std::cout;
using std::endl;


MCPacking::PackingImplementation::PackingImplementation(MCPacking* ptr) : packing_(ptr) {
	numParticles = 0;
	nDim = 0;
	phi = 0;
	simControl.neighborType = simControlStruct::neighborEnum::allToAll;
	simControl.distanceType = simControlStruct::distanceEnum::normal;
	simControl.geometryType = simControlStruct::geometryEnum::normal;
	simControl.potentialType = simControlStruct::potentialEnum::contactPower;
	simControl.constraintType = simControlStruct::constraintEnum::normal;
	simControl.einsteinType = simControlStruct::einsteinEnum::none;
	simControl.particleType = simControlStruct::particleEnum::sphere;
	simControl.monteCarloType = simControlStruct::monteCarloEnum::metropolis;
	simControl.ensembleType = simControlStruct::ensembleEnum::nvt;
}

double MCPacking::PackingImplementation::calcVolumePrefactor(int dim){
	return pow(PI,double(dim)/2)/tgamma(double(dim)/2 + 1);
}

void MCPacking::PackingImplementation::rescaleForces(double maxForceRescale){
	double maxElement = *std::max_element(netForce.begin(), netForce.end());
	if (maxElement > maxForceRescale){
		double rescaleValue = maxForceRescale / maxElement;
		// #pragma omp parallel for
		for (long i = 0; i < numParticles * nDim; i++){
			netForce[i] *= rescaleValue;
		}
	}
}

void MCPacking::PackingImplementation::calcCurrentPhi(){
	double radPower = 0.0;
	for (long i = 0; i < numParticles; i++){
		radPower += pow(radii[i],nDim);
	}
	switch(simControl.particleType) {
		case simControlStruct::particleEnum::cube:
			phi = radPower * pow(2.0,nDim) / volume;
			break;
		case simControlStruct::particleEnum::sphere:			
			phi = radPower * calcVolumePrefactor(nDim) / volume;
		break;
	}
}

void MCPacking::PackingImplementation::translateParticle(const long particleID, const vector<double> translation){
	distTravelledSq[particleID] = 0.0;
	switch(simControl.geometryType) {
		case simControlStruct::geometryEnum::nonSquare:
			for (int i = 0; i < nDim; i++){
				pos[particleID * nDim + i] += translation[i];
				unwrappedPos[particleID * nDim + i] += translation[i];
				pos[particleID * nDim + i] -= boxSize[i]*floor(pos[particleID * nDim + i]/boxSize[i]);
				totalDisplacement[particleID * nDim + i] += translation[i];
				distTravelledSq[particleID] += totalDisplacement[particleID * nDim + i]*totalDisplacement[particleID * nDim + i];
			}
			break;
		case simControlStruct::geometryEnum::E8:{
				vector<double> thisPos(nDim);
				vector<int32_t> mirror(nDim,0);
				for (int dimIndex = 0; dimIndex < nDim; dimIndex++){
					thisPos[dimIndex] = pos[particleID * nDim + dimIndex] + translation[dimIndex];
				}
				packing_->e8LatticePoint(thisPos.begin(),mirror);
				for (int dimIndex = 0; dimIndex < nDim; dimIndex++){
					pos[particleID * nDim + dimIndex] = thisPos[dimIndex];
					unwrappedPos[particleID * nDim + dimIndex] += translation[dimIndex];
					totalDisplacement[particleID * nDim + dimIndex] += translation[dimIndex];
					distTravelledSq[particleID] += totalDisplacement[particleID * nDim + dimIndex]*totalDisplacement[particleID * nDim + dimIndex];
				}
			}
			break;
		case simControlStruct::geometryEnum::Dn:{
				vector<double> thisPos(nDim);
				vector<int32_t> mirror(nDim,0);
				for (int dimIndex = 0; dimIndex < nDim; dimIndex++){
					thisPos[dimIndex] = pos[particleID * nDim + dimIndex] + translation[dimIndex];
				}
				packing_->dnLatticePoint(thisPos.begin(),mirror);
				for (int dimIndex = 0; dimIndex < nDim; dimIndex++){
					pos[particleID * nDim + dimIndex] = thisPos[dimIndex];
					unwrappedPos[particleID * nDim + dimIndex] += translation[dimIndex];
					totalDisplacement[particleID * nDim + dimIndex] += translation[dimIndex];
					distTravelledSq[particleID] += totalDisplacement[particleID * nDim + dimIndex]*totalDisplacement[particleID * nDim + dimIndex];
				}
			}
			break;
		case simControlStruct::geometryEnum::leesEdwards:{
			double x_floor = 0.0;
			for (int i = 0; i < nDim; i++){
				unwrappedPos[particleID * nDim + i] += translation[i];
				totalDisplacement[particleID * nDim + i] += translation[i];
				distTravelledSq[particleID] += totalDisplacement[particleID * nDim + i]*totalDisplacement[particleID * nDim + i];
				pos[particleID * nDim + i] += translation[i];
				switch (i){
					case 0:
						x_floor = floor(pos[particleID * nDim + i]);
						pos[particleID * nDim + i] -= x_floor;
						break;
					case 1:
						pos[particleID * nDim + i] -= LEshift*x_floor;
						pos[particleID * nDim + i] -= floor(pos[particleID * nDim + i]);
						break;
					default:
						pos[particleID * nDim + i] -= floor(pos[particleID * nDim + i]);
				}
			}
			break;
		}
		default:
			for (int i = 0; i < nDim; i++){
				pos[particleID * nDim + i] += translation[i];
				unwrappedPos[particleID * nDim + i] += translation[i];
				pos[particleID * nDim + i] -= floor(pos[particleID * nDim + i]);
				totalDisplacement[particleID * nDim + i] += translation[i];
				distTravelledSq[particleID] += totalDisplacement[particleID * nDim + i]*totalDisplacement[particleID * nDim + i];
			}
			break;
	}
}

// Converts a padded gpu boolean matrix to an unpadded csr array on the device
void MCPacking::PackingImplementation::paddedToCSR(vector<long> &paddedIndices, vector<long> &rowSizes, vector<long> &indices, vector<long> &indptr) {
	indptr.resize(rowSizes.size() + 1);
	indptr[0] = 0;
	// std::inclusive_scan(rowSizes.begin(), rowSizes.end(), indptr.begin() + 1);
	for (size_t i = 0; i < rowSizes.size(); i++){
		indptr[i+1] = rowSizes[i] + indptr[i];
	}
	indices.resize(indptr[rowSizes.size()]);
	copy_if(paddedIndices.begin(), paddedIndices.end(), indices.begin(), [](long i){return !(i<0);});
}

void MCPacking::PackingImplementation::paddedToCOO(vector<long> &paddedIndices, vector<long> &rowSizes, vector<double> &paddedData, vector<long> &indices1, vector<long> &indices2, vector<double> &data, long blockSize1, long blockSize2) {
	vector<long> indptr(rowSizes.size() + 1);
	indptr[0] = 0;
	// thrust::inclusive_scan(rowSizes.begin(), rowSizes.end(), indptr.begin() + 1);
	for (size_t i = 0; i < rowSizes.size(); i++){
		indptr[i+1] = rowSizes[i] + indptr[i];
	}
	indices1.resize(indptr[rowSizes.size()]);
	indices2.resize(indptr[rowSizes.size()]);
	data.resize(indptr[rowSizes.size()] * blockSize1 * blockSize2);
	long indexPad = paddedIndices.size() / rowSizes.size();
	long dataPad = paddedData.size() / (rowSizes.size() * blockSize1);
	for(size_t i = 0; i < rowSizes.size(); i++){
		long start = indptr[i];
		long rowSize = indptr[i+1] - start;
		for (long j = 0; j < indexPad; j++) {
			if (j < rowSize) {
				indices1[start + j] = i;
				indices2[start + j] = paddedIndices[i * indexPad + j];
				for(long k = 0; k < blockSize1; k++) {
					for(long l=0; l < blockSize2; l++) {
						data[((start + j) * blockSize1 + k) * blockSize2 + l] = paddedData[(i * blockSize1 + k) * dataPad + j * blockSize2 + l];
					}
				}
			}
		}
	}
}

void MCPacking::PackingImplementation::doubleVectorToCSR(vector< vector<long> > &doubleVector, vector<long> &indices, vector<long> &indptr) {
	for (size_t i = 0; i < doubleVector.size(); i++){
		for (size_t j = 0; j < doubleVector[i].size(); j++){
			indices.push_back(doubleVector[i][j]);
		}
	}
	indptr.resize(doubleVector.size() + 1);
	indptr[0] = 0;
	// std::inclusive_scan(rowSizes.begin(), rowSizes.end(), indptr.begin() + 1);
	for (size_t i = 0; i < doubleVector.size(); i++){
		indptr[i+1] = doubleVector[i].size() + indptr[i];
	}
}

void MCPacking::PackingImplementation::doubleVectorToCOO(vector< vector<long> >& doubleVectorIndices, vector< vector<double> >& doubleVectorData, vector<long> &indices1, vector<long> &indices2, vector<double> &data, long blockSize1, long blockSize2) {
	for(size_t i = 0; i < doubleVectorIndices.size(); i++){
		for (size_t j = 0; j < doubleVectorIndices[i].size(); j++) {
			indices1.push_back(i);
			indices2.push_back(doubleVectorIndices[i][j]);
			for(long k = 0; k < blockSize1; k++) {
				for(long l=0; l < blockSize2; l++) {
					data.push_back(doubleVectorData[i][j * blockSize2 * blockSize1 + k * blockSize1 + l]);
					// data.push_back(doubleVectorData[i][k * blockSize1 + j * blockSize2 + l]);
				}
			}
		}
	}
}

// Converts a padded gpu block matrix with leading diagonal to an unpadded block coo on the device
// this format is used to store the Hessian, where diagonal blocks ead each row
// void MCPacking::PackingImplementation::diagonalPaddedToCOO(vector<long> &paddedIndices, vector<long> &rowSizes, vector<double> &paddedData, vector<long> &indices1, vector<long> &indices2, vector<double> &data, long blockSize1, long blockSize2) {
// 	vector<long> indptr(rowSizes.size() + 1);
// 	indptr[0] = 0;
// 	for (size_t i = 0; i < rowSizes.size(); i++){
// 		indptr[i+1] = rowSizes[i] + indptr[i];
// 	}
// 	//to store diagonal intries, one extra element is included in each row
// 	thrust::transform(indptr.begin(), indptr.end(), thrust::counting_iterator<long>(0), indptr.begin(), thrust::plus<long>());

// 	indices1.resize(indptr[rowSizes.size()]);
// 	indices2.resize(indptr[rowSizes.size()]);
// 	data.resize(indptr[rowSizes.size()] * blockSize1 * blockSize2);

// 	long indexPad = paddedIndices.size() / rowSizes.size();
// 	long dataPad = paddedData.size() / (rowSizes.size() * blockSize1);

// 	long* paddedIndicesPtr = thrust::raw_pointer_cast(&paddedIndices[0]);
// 	devnum* paddedDataPtr = thrust::raw_pointer_cast(&paddedData[0]);
// 	long* indices1Ptr = thrust::raw_pointer_cast(&indices1[0]);
// 	long* indices2Ptr = thrust::raw_pointer_cast(&indices2[0]);
// 	devnum* dataPtr = thrust::raw_pointer_cast(&data[0]);
// 	long* indptrPtr = thrust::raw_pointer_cast(&indptr[0]);
// 	for (long i = 0; i < rowSizes.size(); i++) {
// 		long start = indptrPtr[i];
// 		indices1Ptr[start] = i;
// 		indices2Ptr[start] = i;
// 		for(long k = 0; k < blockSize1; k++) {
// 			for(long l = 0; l < blockSize2; l++) {
// 				dataPtr[(start * blockSize1 + k) * blockSize2 + l] = paddedDataPtr[(i * blockSize1 + k) * dataPad + l];
// 			}
// 		}
// 		start++;
// 		long rowSize = indptrPtr[i+1] - start;
// 		for (long j = 0; j < indexPad; j++) {
// 			if (j < rowSize) {
// 				indices1Ptr[start + j] = i;
// 				indices2Ptr[start + j] = paddedIndicesPtr[i * indexPad + j];
// 				for(long k = 0; k < blockSize1; k++) {
// 					for(long l = 0; l < blockSize2; l++) {
// 						dataPtr[((start + j) * blockSize1 + k) * blockSize2 + l] = paddedDataPtr[(i * blockSize1 + k) * dataPad + (j + 1) * blockSize2 + l];
// 					}
// 				}
// 			}
// 		}
// 	}
// }

//Calculate and return the list of gaps below a threshhold
void MCPacking::PackingImplementation::calcSmallGaps(vector< vector<double> >& gapList, vector< vector<long> >& gapIndices, double maxGap){
	for (long pID = 0; pID < numParticles; pID++) {
		// #pragma omp parallel for reduction(+: numAddedGaps)
		for(size_t i = 0; i < packing_->neighborList[pID].size(); i++){
			long nID = packing_->neighborList[pID][i];
			double gap = packing_->calcGap(pID, nID);
			if (gap < maxGap) {
				gapIndices[pID].push_back(nID);
				gapList[pID].push_back(gap);
			}
		}
	}
}

void MCPacking::PackingImplementation::calcSmallOverlaps(vector< vector<double> >& overlapList, vector< vector<long> >& overlapIndices, double minOverlap){
	for (long pID = 0; pID < numParticles; pID++) {
		// #pragma omp parallel for reduction(+: numAddedGaps)
		for(size_t i = 0; i < packing_->neighborList[pID].size(); i++){
			long nID = packing_->neighborList[pID][i];
			double overlap = packing_->calcOverlap(unwrappedPos.begin() + pID * nDim, unwrappedPos.begin() + nID * nDim, pID, nID);
			if (overlap > minOverlap) {
				overlapIndices[pID].push_back(nID);
				overlapList[pID].push_back(overlap);
			}
		}
	}
}

struct saxpy_functor{
    const double a;
    saxpy_functor(double _a) : a(_a) {}
    double operator()(const double& x, const double& y) const {
        return a * x + y;
    }
};

void saxpy_fast(double A, vector<double>& X, vector<double>& Y){
    // Y <- A * X + Y
    transform(X.begin(), X.end(), Y.begin(), Y.begin(), saxpy_functor(A));
}

double MCPacking::PackingImplementation::projectVector(vector<double>& vectorIn) {
	double vectorDotConstraint = inner_product(vectorIn.begin(), vectorIn.end(), constraint.begin(), 0.0);
	saxpy_fast(-vectorDotConstraint, constraint, vectorIn);
	return vectorDotConstraint;
}

#endif /* PACKINGIMPtempEMENTATION_H_ */